
;*****************************************************;
; dosname.asm                                         ;
; -- VxD that returns the name of the currently       ;
;    executing DOS app in a given VM.                 ;
;*****************************************************;
.386p

; Header files.
include vmm.inc
include v86mmgr.inc

; Constants.
wDeviceId     equ 297bH ; Device id from Microsoft.
bVersionMajor equ 01    ; Major byte of version number.
bVersionMinor equ 00    ; Minor byte of version number.
wVersion      equ (bVersionMajor * 256 + bVersionMinor)
cchPathMax    equ 128   ; Max. size of pathname.
ibEnv         equ 2ch   ; Env offset in PSP.

; API id's.
apiGetVersion equ 0     ; Get version of this VxD.
apiDosName    equ 1     ; Get the name of the DOS app.
apiRegister   equ 2     ; Register a callback function.

; Globals.
VxD_DATA_SEG
selClient       dw  (0)   ; Client buffer selector.
ibClient        dd  (?)   ; Client buffer offset.
hwndClient      dw  (?)   ; Window to notify.
pfnNotifySel    dw  (?)   ; Code selector of callback.
pfnNotifyOffset dw  (?)   ; Offset in selector.
rgchPath        db  (cchPathMax + 1) dup (?)
VxD_DATA_ENDS

; Device Header.
Declare_Virtual_Device  dosname, bVersionMajor,      \
    bVersionMinor, DosNameControl, wDeviceId,        \
    Undefined_Init_Order,, DosNameApi

; Control Procedures.
VxD_Locked_Code_Seg

BeginProc   DosNameInit
;*****************************************************;
; -- Called for protect mode initialization.  Not     ;
;    much to do.                                      ;
;*****************************************************;
    clc
    ret
EndProc     DosNameInit

BeginProc    DosNameControl
;*****************************************************;
; -- Control dispatcher.  Not much to do here either. ;
;*****************************************************;
    Control_Dispatch    Device_Init, DosNameInit
    clc
    ret
EndProc DosNameControl
VxD_Locked_Code_Ends

; API's.
VxD_Code_Seg

BeginProc   DosNameApi, Public
;*****************************************************;
; -- Main dispatch point for protected mode API.      ;
; -- On input:                                        ;
;    -- Client_EAX : API id to execute.               ;
; -- Jump to corresponding API routine.               ;
;*****************************************************;
    ; Innocent until proven guilty (no error yet).
    mov         [ebp].Client_Flags, 0

    ; Which function is required?
    movzx       eax, [ebp].Client_AX

    cmp         ax, apiGetVersion
    jz short    DosNameGetVersion

    cmp         ax, apiDosName
    jz short    DosNameTest

    cmp         ax, apiRegister
    jz          RegisterClient

    ; Guilty.  Set carry flag to indicate failure.
    or          [ebp].Client_Flags, CF_Mask
    ret
EndProc     DosNameApi

;*****************************************************;
; -- Return the version number.                       ;
;*****************************************************;
BeginProc   DosNameGetVersion, Public
    mov         [ebp].Client_AX, wVersion
    ret
EndProc     DosNameGetVersion

;*****************************************************;
; -- Return the name of the currently executing DOS   ;
;    task in the given VM.                            ;
; -- Schedules a callback for the VM, so that it can  ;
;    be examined when it is running (and in memory).  ;
; -- On Entry:                                        ;
;    -- Client_BX, : Low word of VM handle.           ;
;    -- Client_CX, : High word of VM handle.          ;
;    -- Client_DS, : Buffer to receive name of        ;
;       Client_DX  : running DOS app.                 ;
;    -- Client_SI, : Window to receive notification.  ;
;*****************************************************;
BeginProc   DosNameTest, Public
    ; Make sure we aren't already processing a request.
    test        selClient, 0ffffh
    jnz short   DosNameTestExit

    ; Get VM from caller.
    mov         bx, [ebp].Client_CX
    shl         ebx, 10h
    mov         bx, [ebp].Client_BX

    ; Make sure VM handle is valid.
    VMMCall     Validate_VM_Handle
    jc short    DosNameTestExit

    ; Get window to receive notification.
    mov         ax, [ebp].Client_SI
    mov         [hwndClient], ax

    ; Get a selector for VM's memory.
    VxDCall     V86MMGR_Get_VM_Flat_Sel
    or          ax, ax
    jz short    DosNameTestExit
    movzx       edx, ax     ; Flat VM sel as ref data.

    ; Store the buffer to receive DOS app path.
    mov         ax, [ebp].Client_DS
    mov         [selClient], ax
    movzx       eax, [ebp].Client_DX
    mov         [ibClient], eax

    ; Schedule the callback that does the work.  We
    ; already have the handle of the VM in EBX and a
    ; flat selector to the VM's memory in EDX.
    mov         eax, High_Pri_Device_Boost  ; Now!
    mov         ecx, PEF_Wait_For_STI or \
                     PEF_Wait_Not_Crit
    mov         esi, OFFSET32 GetDosAppName
    VMMCall     Call_Priority_VM_Event

    ; If VM did not execute, temporarily set its
    ; execution focus.
    test        esi, 0ffffffffh
    jz short    DosNameTestExit
    VMMCall     Set_Execution_Focus

DosNameTestExit:
    ret
EndProc     DosNameTest

;*****************************************************;
; -- Call_Priority_VM_Event callback.                 ;
; -- Issues an int 21, service 6200 to get the        ;
;    current PSP, then copies the app name found at   ;
;    the end of the environment block.                ;
; -- On entry:                                        ;
;    -- EBX : VM being examined.                      ;
;    -- EDX : Loword has flat selector for this VM.   ;
;*****************************************************;
BeginProc       GetDosAppName, Public
    push        edx

    ; Get the current PSP for this VM.
    Push_Client_State
    VMMCall     Begin_Nest_Exec
    mov         [ebp].Client_AX, 6200h
    mov         eax, 21h
    VMMCall     Exec_Int
    VMMCall     End_Nest_Exec
    movzx       eax, WORD PTR [ebp].Client_BX
    shl         eax, 4
    add         eax, ibEnv
    Pop_Client_State

    ; Find environment block and walk to end.
    mov         [rgchPath], 0
    pop         edx
    push        ds
    pop         es
    mov         ds, dx
    movzx       eax, WORD PTR [eax]
    shl         eax, 4
    mov         edi, eax

EnvWalk:
    cmp         byte ptr [edi], 0
    jne short   SkipEnvLine
    cmp         byte ptr [edi + 1], 1
    jne short   ScheduleNotify  ; No app name found.
    cmp         byte ptr [edi + 2], 0
    jne short   ScheduleNotify  ; No app name found.
    add         edi, 3
    mov         esi, edi

    ; Copy app path to global.
    mov         edi, OFFSET32 rgchPath
    mov         ecx, cchPathMax
    repz movsb

ScheduleNotify:
    ; Schedule call back to system VM to call client's
    ; notification callback.
    push        es
    pop         ds
    mov         edx, ebx    ; Current VM handle param.
    VMMCall     Get_Sys_VM_Handle
    mov         eax, High_Pri_Device_Boost
    mov         ecx, PEF_Wait_For_STI or \
                     PEF_Wait_Not_Crit
    mov         esi, offset32 NotifyClient
    VMMCall     Call_Priority_VM_Event
    jmp short   GotName2

SkipEnvLine:
    ; Skip characters until null.
    mov       ecx, cchPathMax
    mov       al, 0
    repne scasb
    or        ecx, ecx
    jnz short EnvWalk ; Make sure there is a null.

GotName1:
    push      es
    pop       ds

GotName2:
    ret
EndProc     GetDosAppName

;*****************************************************;
; -- Call_Priority_VM_Event callback.                 ;
; -- Call the client callback to notify that the path ;
;    has been obtained.                               ;
; -- On entry, EDX VM handle containing DOS app.      ;
;*****************************************************;
BeginProc   NotifyClient, Public
    ; Copy app path to client's buffer.
    xor         ax, ax
    mov         al, BYTE PTR [rgchPath]
    or          ax, ax
    jz short    CallBack
    mov         edi, [ibClient]
    mov         esi, OFFSET32 rgchPath
    mov         ax, [selClient]
    push        es
    mov         es, ax
    mov         ecx, cchPathMax
    repz movsb
    pop         es

CallBack:
    ; Call client notification callback.
    Push_Client_State
    VMMCall     Begin_Nest_Exec

    ; Push VM handle.
    mov         eax, edx
    shr         eax, 10h
    VMMCall     Simulate_Push
    mov         ax, dx
    VMMCall     Simulate_Push

    ; Push notification window.  Make sure we have one.
    mov         ax, [hwndClient]
    or          ax, ax
    jz          short NotifyClientExit
    VMMCall     Simulate_Push

    mov         cx, [pfnNotifySel]
    movzx       edx, [pfnNotifyOffset]
    VMMCall     Simulate_Far_Call
    VMMCall     Resume_Exec

NotifyClientExit:
    VMMCall     End_Nest_Exec
    Pop_Client_State

    ; Reset for next time.
    mov         [selClient], 0
    ret
EndProc     NotifyClient

;*****************************************************;
; -- Register the client's callback function.         ;
;    -- Client_DX, : Pointer to callback to invoke    ;
;       Client_BX  : when path obtained.              ;
;*****************************************************;
BeginProc   RegisterClient, Public
    ; Store address of notification callback.
    mov         ax, [ebp].Client_BX     ; Offset.
    mov         [pfnNotifyOffset], ax
    mov         ax, [ebp].Client_DX     ; Selector.
    mov         [pfnNotifySel], ax
    ret
EndProc     RegisterClient
VxD_Code_Ends
End
------------------------------------------------------------
	Paul Bonneau
	bonneau@hyper.hyper.com
	(519) 725-4040

