#include <windows.h>
#include "msgboxex.h"

#define    MSGBOXEX_SPACING         18
#define    MSGBOXEX_BUTTONWIDTH     58
#define    MSGBOXEX_BUTTONHEIGHT    28
#define    MSGBOXEX_BUTTONSPACING   16
#define    MSGBOXEX_BOTTOMMARGIN    10
#define    MSGBOXEX_MAXWIDTH        240
#define    MSGBOXEX_MINHEIGHT       70

#define    IDD_ICON                 0xA90

typedef struct {
    long    dtStyle;
    BYTE    dtItemCount;
    int    dtX;
    int    dtY;
    int    dtCX;
    int    dtCY;
    char    dtMenuName[ 1 ];
    char    dtClassName[ 1 ];
    char    dtCaptionText[ 1 ];
} DLGTEMPLATE;

typedef struct {
    int    dtilX;
    int    dtilY;
    int    dtilCX;
    int    dtilCY;
    int    dtilID;
    long    dtilStyle;
    char    dtilClass[ 1 ];
    char    dtilText[ 1 ];
    BYTE    dtilInfo;
} DLGITEMTEMPLATE;

static int AddDlgItem( HANDLE, int, int, int, int,
                              int, int, long, LPSTR, LPSTR );
static int AddButtonItem( HANDLE, int, int, int,
                                  int, LPSTR );

WNDPROC lpfnHookProc = NULL;    /* MessageBox hook
                                            procedure */
HICON hIcon;                        /* Handle of icon */
int nDefButton;                    /* Default button */

BOOL __export FAR PASCAL MsgBoxDlgProc( HWND,
    unsigned, WORD, LONG );

LONG __export FAR PASCAL IconProc( HWND, unsigned,
    WORD, LONG );

char * szOK     = "OK";
char * szCancel = "Cancel";
char * szIgnore = "&Ignore";
char * szRetry  = "&Retry";
char * szAbort  = "&Abort";
char * szHelp   = "&Help";
char * szYes    = "&Yes";
char * szNo     = "&No";

/* MessageBoxEx  */

int FAR PASCAL MessageBoxEx(
    HWND hwnd, LPSTR lpcszStr,
    LPSTR lpcszTitle, unsigned fuStyle,
    unsigned fuStyleEx, WNDPROC lpfMsgProc )
{
    DLGTEMPLATE FAR * lpDlgTmp;
    HINSTANCE hInst;
    FARPROC lpMsgBoxDlgProc;
    HGLOBAL hDlgTmp;
    int nRetCode, cwButtonRow, cbTemplate, cxMsgBox, cyMsgBox;
    int cItems, dxLeft, x, y;
    HDC hdc;
    RECT rc;

    if( fuStyle == (MB_ICONHAND|MB_SYSTEMMODAL) &&
         fuStyleEx == 0 )
        return( MessageBox( hwnd, lpcszStr, lpcszTitle,
                                  fuStyle ) );

    hdc = CreateDC( "DISPLAY", NULL, NULL, NULL );
    SelectObject( hdc, GetStockObject( SYSTEM_FONT ) );
    SetRect( &rc, 0, 0, MSGBOXEX_MAXWIDTH, 0 );
    DrawText( hdc, lpcszStr, -1, &rc, DT_LEFT|
                 DT_NOPREFIX|DT_WORDBREAK|DT_CALCRECT );
    rc.right += 10;
    DeleteDC( hdc );

    if( !lpcszTitle )
        lpcszTitle = "Error";

    cyMsgBox = rc.bottom + ( MSGBOXEX_SPACING * 2 );
    cxMsgBox = rc.right + ( MSGBOXEX_SPACING * 2 );
    dxLeft = MSGBOXEX_SPACING;
    hIcon = NULL;

    cbTemplate = sizeof(DLGTEMPLATE) + lstrlen(lpcszTitle);
    hDlgTmp = GlobalAlloc( GHND, cbTemplate );
    if( !hDlgTmp )
        return( 0 );

    lpDlgTmp = (DLGTEMPLATE FAR *)GlobalLock( hDlgTmp );
    if( !lpDlgTmp ) {
        GlobalFree( hDlgTmp );
        return( 0 );
        }

    lpDlgTmp->dtStyle = DS_MODALFRAME | WS_POPUP |
                              WS_CAPTION | WS_SYSMENU;
    lpDlgTmp->dtItemCount = 0;
    lpDlgTmp->dtX = 0;
    lpDlgTmp->dtY = 0;
    lpDlgTmp->dtMenuName[ 0 ] = '\0';
    lpDlgTmp->dtClassName[ 0 ] = '\0';
    lstrcpy( lpDlgTmp->dtCaptionText, lpcszTitle );
    GlobalUnlock( hDlgTmp );

    if( fuStyle & MB_SYSTEMMODAL )
        lpDlgTmp->dtStyle |= DS_SYSMODAL;

    if( cyMsgBox < MSGBOXEX_MINHEIGHT )
        cyMsgBox = MSGBOXEX_MINHEIGHT;

    if( fuStyle & MB_ICONMASK ) {
        int cxIcon;
        int cyIcon;
        LPSTR lpIcon;
        switch( fuStyle & MB_ICONMASK ) {
            case MB_ICONEXCLAMATION:
                lpIcon = (LPSTR)IDI_EXCLAMATION;
                break;
            case MB_ICONHAND:
                lpIcon = (LPSTR)IDI_HAND;
                break;
            case MB_ICONQUESTION:
                lpIcon = (LPSTR)IDI_QUESTION;
                break;
            case MB_ICONASTERISK:
                lpIcon = (LPSTR)IDI_ASTERISK;
                break;
            }
        hIcon = LoadIcon( NULL, lpIcon );

        cxIcon = GetSystemMetrics( SM_CXICON );
        cyIcon = GetSystemMetrics( SM_CYICON );
        dxLeft += MSGBOXEX_SPACING + cxIcon;
        cxMsgBox += dxLeft;
        y = ( cyMsgBox - cyIcon ) / 2;

        cbTemplate += AddDlgItem( hDlgTmp, cbTemplate,
                MSGBOXEX_SPACING, y, cxIcon, cyIcon,
                IDD_ICON, WS_CHILD|SS_ICON,"static", "" );
        }
    y = ( cyMsgBox - rc.bottom ) / 2;
    cbTemplate += AddDlgItem( hDlgTmp, cbTemplate,
                dxLeft, y, rc.right, rc.bottom, -1,
                WS_CHILD|SS_LEFT,"static", lpcszStr );
    switch( fuStyle & MB_TYPEMASK ) {
        case MB_OK:                 cItems = 1;    break;
        case MB_OKCANCEL:           cItems = 2;    break;
        case MB_YESNO:              cItems = 2;    break;
        case MB_YESNOCANCEL:        cItems = 3;    break;
        case MB_ABORTRETRYIGNORE:   cItems = 3;    break;
        case MB_RETRYCANCEL:        cItems = 2;    break;
        }
    if( fuStyleEx & MB_EX_HELP )
        ++cItems;
    cwButtonRow = MSGBOXEX_BUTTONWIDTH * cItems +
            ( MSGBOXEX_BUTTONSPACING * ( cItems - 1 ) );
    if( cwButtonRow > ( cxMsgBox - dxLeft ) )
        cxMsgBox = dxLeft + cwButtonRow +
                      MSGBOXEX_BUTTONSPACING * 2;
    lpfnHookProc = lpfMsgProc;
    nDefButton = 1;
    switch( fuStyle & MB_DEFMASK ) {
        case MB_DEFBUTTON1:    nDefButton = 1;    break;
        case MB_DEFBUTTON2:    nDefButton = 2;    break;
        case MB_DEFBUTTON3:    nDefButton = 3;    break;
        }
    if( fuStyleEx & MB_EX_DEFBUTTON4 )
        nDefButton = 4;
    x = ( cxMsgBox - cwButtonRow ) / 2;
    switch( fuStyle & MB_TYPEMASK ) {
        case MB_OK:
            cbTemplate += AddButtonItem( hDlgTmp,
                    cbTemplate, x, cyMsgBox, IDOK, szOK );
            break;
        case MB_OKCANCEL:
            cbTemplate += AddButtonItem( hDlgTmp,
                    cbTemplate, x, cyMsgBox, IDOK, szOK );
            cbTemplate += AddButtonItem( hDlgTmp,
                    cbTemplate, x += MSGBOXEX_BUTTONSPACING +
                    MSGBOXEX_BUTTONWIDTH, cyMsgBox, IDCANCEL,
                    szCancel );
            break;
        case MB_YESNO:
            cbTemplate += AddButtonItem( hDlgTmp,
                    cbTemplate, x, cyMsgBox, IDYES, szYes );
            cbTemplate += AddButtonItem( hDlgTmp,
                    cbTemplate, x += MSGBOXEX_BUTTONSPACING +
                    MSGBOXEX_BUTTONWIDTH, cyMsgBox, IDNO,
                    szNo );
            break;
        case MB_YESNOCANCEL:
            cbTemplate += AddButtonItem( hDlgTmp,
                    cbTemplate, x, cyMsgBox, IDYES, szYes );
            cbTemplate += AddButtonItem( hDlgTmp,
                    cbTemplate, x += MSGBOXEX_BUTTONSPACING +
                    MSGBOXEX_BUTTONWIDTH, cyMsgBox, IDNO, szNo );
            cbTemplate += AddButtonItem( hDlgTmp,
                    cbTemplate, x += MSGBOXEX_BUTTONSPACING +
                    MSGBOXEX_BUTTONWIDTH, cyMsgBox, IDCANCEL,
                    szCancel );
            break;
        case MB_ABORTRETRYIGNORE:
            cbTemplate += AddButtonItem( hDlgTmp,
                    cbTemplate, x, cyMsgBox, IDABORT,
                    szAbort );
            cbTemplate += AddButtonItem( hDlgTmp,
                    cbTemplate, x += MSGBOXEX_BUTTONSPACING +
                    MSGBOXEX_BUTTONWIDTH, cyMsgBox, IDRETRY,
                    szRetry );
            cbTemplate += AddButtonItem( hDlgTmp,
                    cbTemplate, x += MSGBOXEX_BUTTONSPACING +
                    MSGBOXEX_BUTTONWIDTH, cyMsgBox, IDIGNORE,
                    szIgnore );
            break;
        case MB_RETRYCANCEL:
            cbTemplate += AddButtonItem( hDlgTmp,
                    cbTemplate, x, cyMsgBox, IDRETRY,
                    szRetry );
            cbTemplate += AddButtonItem( hDlgTmp,
                    cbTemplate, x += MSGBOXEX_BUTTONSPACING +
                    MSGBOXEX_BUTTONWIDTH, cyMsgBox,
                    IDCANCEL, szCancel );
            break;
        }
    if( fuStyleEx & MB_EX_HELP )
        cbTemplate += AddButtonItem( hDlgTmp,
                    cbTemplate, x += MSGBOXEX_BUTTONSPACING +
                    MSGBOXEX_BUTTONWIDTH, cyMsgBox, IDEXHELP,
                    szHelp );
    lpDlgTmp = (DLGTEMPLATE FAR *)GlobalLock( hDlgTmp );
    if( !lpDlgTmp ) {
        GlobalFree( hDlgTmp );
        return( 0 );
        }
    cyMsgBox += MSGBOXEX_BOTTOMMARGIN +
                    MSGBOXEX_BUTTONHEIGHT;
    cyMsgBox = ( cyMsgBox * 8 ) /
                        HIWORD( GetDialogBaseUnits() );
    cxMsgBox = ( cxMsgBox * 4 ) /
                        LOWORD( GetDialogBaseUnits() );
    lpDlgTmp->dtCX = cxMsgBox;
    lpDlgTmp->dtCY = cyMsgBox;
    GlobalUnlock( hDlgTmp );

    hInst = (HINSTANCE)GetWindowWord( hwnd, GWW_HINSTANCE );
    lpMsgBoxDlgProc = MakeProcInstance(
                                (FARPROC)MsgBoxDlgProc,
                                hInst );
    nRetCode = DialogBoxIndirect( hInst, hDlgTmp,
                                hwnd, (DLGPROC)lpMsgBoxDlgProc );
    FreeProcInstance( lpMsgBoxDlgProc );
    GlobalFree( hDlgTmp );
    return( nRetCode );
}


static int AddButtonItem( HANDLE hDlgTmp,
    int cbTemplate, int x, int y,
    int dtilID, LPSTR lpszStr )
{
    long lStyle;

    lStyle = WS_CHILD | WS_TABSTOP |
                ( --nDefButton ? BS_PUSHBUTTON :
                BS_DEFPUSHBUTTON );
    return( AddDlgItem( hDlgTmp, cbTemplate, x, y,
                    MSGBOXEX_BUTTONWIDTH,
                    MSGBOXEX_BUTTONHEIGHT,
                    dtilID, lStyle, "button", lpszStr ) );
}


BOOL FAR PASCAL MsgBoxDlgProc( HWND hwnd,
    unsigned message, WORD wParam, LONG lParam )
{
    if( lpfnHookProc )
        {
        LONG lResult;

        lResult = (*lpfnHookProc)( hwnd, message,
                                           wParam, lParam );
        if( lResult && message != WM_INITDIALOG )
            return( (BOOL)lResult );
        }
    switch( message )
        {
        case WM_INITDIALOG: {
            RECT rc;
            if( hIcon ) {
                HWND hwndIcon;
                hwndIcon = GetDlgItem( hwnd, IDD_ICON );
                SetWindowLong( hwndIcon, GWL_WNDPROC,
                                   (LONG)IconProc );
                }
            GetWindowRect( hwnd, &rc );
            SetWindowPos( hwnd, NULL,
                    ( GetSystemMetrics( SM_CXSCREEN ) -
                        ( rc.right - rc.left ) ) / 2,
                    ( GetSystemMetrics( SM_CYSCREEN ) -
                        ( rc.bottom - rc.top ) ) / 3,
                    0, 0, SWP_NOSIZE | SWP_NOACTIVATE );
            return( TRUE );
            }
        case WM_COMMAND:
            if( wParam != IDEXHELP )
                EndDialog( hwnd, wParam );
            break;
        default:
            return( FALSE );
        }
    return( FALSE );
}


LONG FAR PASCAL IconProc( HWND hwnd, unsigned message,
    WORD wParam, LONG lParam )
{
    if( message == WM_PAINT ) {
        PAINTSTRUCT ps;
        HDC hdc;

        hdc = BeginPaint( hwnd, &ps );
        DrawIcon( hdc, 0, 0, hIcon );
        EndPaint( hwnd, &ps );
        }
    return( FALSE );
}

static int AddDlgItem( HANDLE hDlgTmp, int cbTemplate,
    int dtilX, int dtilY, int dtilCX, int dtilCY,
    int dtilID, long dtilStyle, LPSTR lpdtilClass,
    LPSTR lpdtilText )
{
    DLGTEMPLATE FAR * lpDlgTmp;
    DLGITEMTEMPLATE FAR * lpDlgItemTmp;
    int cbItem;
    LPSTR lpstr;

    cbItem = sizeof( DLGITEMTEMPLATE ) +
                lstrlen( lpdtilClass ) +
                lstrlen( lpdtilText );
    hDlgTmp = GlobalReAlloc( hDlgTmp, cbTemplate +
                cbItem, GMEM_MOVEABLE );
    lpDlgTmp = (DLGTEMPLATE FAR *)GlobalLock( hDlgTmp );
    ++lpDlgTmp->dtItemCount;
    dtilY = ( dtilY * 8 ) / HIWORD( GetDialogBaseUnits() );
    dtilX = ( dtilX * 4 ) / LOWORD( GetDialogBaseUnits() );
    dtilCY = ( dtilCY * 8 ) / HIWORD( GetDialogBaseUnits() );
    dtilCX = ( dtilCX * 4 ) / LOWORD( GetDialogBaseUnits() );
    lpDlgItemTmp = (DLGITEMTEMPLATE FAR *)(
                            (LPSTR)lpDlgTmp + cbTemplate );
    lpDlgItemTmp->dtilX = dtilX;
    lpDlgItemTmp->dtilY = dtilY;
    lpDlgItemTmp->dtilCX = dtilCX;
    lpDlgItemTmp->dtilCY = dtilCY;
    lpDlgItemTmp->dtilID = dtilID;
    lpDlgItemTmp->dtilStyle = dtilStyle | WS_VISIBLE;

    /* Copy in the class name. */
    lpstr = (LPSTR)&lpDlgItemTmp->dtilClass;
    lstrcpy( lpstr, lpdtilClass );

    /* Copy in the item text */
    lpstr += lstrlen( lpdtilClass ) + 1;
    lstrcpy( lpstr, lpdtilText );

    lpstr += lstrlen( lpdtilText ) + 1;
    *lpstr = 0;
    GlobalUnlock( hDlgTmp );
    return( cbItem );
}
