/*  SWITCH.C                                        Donna Campanella
    Program to switch between Monochrome mode and EGA/VGA Color Mode
    for computers with both a color and monochrome video adapter and
    dual monitors.  Detects the currently active display and switches
    to the alternate one.  Does not clear the screen when switching
    to color mode as DOS's MODE.COM does.                             */

#include <dos.h>
#define MONO_MODE          0x30
#define MONO               7
#define COLOR              3
#define VIDEO              0x10
#define SET_VIDEO_MODE     0x00
#define SET_CURSOR         0x02
#define CURSOR_COL         0
#define SCREEN_ROW         0x18
#define VIDEO_PAGE         0
#define CURRENT_MODE_ADDR  0x410

/*   CURRENT_MODE_ADDR:  The two-byte word at 410h as defined in
     Norton's Programmer's Guide to the IBM PC:      (pg 52-53)
     Word that holds the equipment list that is reported
     by equipment-list service interrupt 17 (hex 11) (pg 219)
     Initial video mode:   7  6  5  4  3  2  1  0
                               x  x
     where: 10 ==> 80 column color
            11 ==> 80 column monochrome
     Set/clear bit to switch modes.                                     */

void main()
{
    union REGS regs;
    int new_mode;
    unsigned char video_mode, far *vid_mode_ptr;

    /* Get current active display */
    vid_mode_ptr = (int far *) CURRENT_MODE_ADDR;
    video_mode = *(vid_mode_ptr);

    /* Test current mode and switch to alternate mode */
    if ((video_mode & MONO_MODE) == MONO_MODE)
    {
        /* clear bit to set color mode */
        video_mode &= 0xEF;
        new_mode = COLOR;
        }
    else
    {
        /* set bit to set monochrome mode */
        video_mode |= 0x10;
        new_mode = MONO;
      }

    /* Enable the display by updating the "equipment list"  */
    *vid_mode_ptr = video_mode;

    /* Switch mode, setting high bit in AL (OR mode with    */
    /* hex 80)  to avoid clearing the EGA/VGA screen.       */
    regs.h.al = (char) (new_mode | 0x80);
    regs.h.ah = SET_VIDEO_MODE;
    int86(VIDEO, &regs, &regs);

    /* Set cursor to bottom of screen */
    regs.h.ah = SET_CURSOR;
    regs.h.dh = SCREEN_ROW;
    regs.h.bh = VIDEO_PAGE;
    regs.h.dl = CURSOR_COL;
    int86(VIDEO, &regs, &regs);
}
